#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"

static void
usage(void)
{
	errx(1, "usage: randcounts [-d delimstr] [-n] [-N num] [-p prec]"
	        "[-r repeats] [-R] [-s seed] weight1 [weight2 ...]");
}

int
main(int argc, char *argv[])
{
	int i, ch, nbins, prec = 17, finalnl = 1, s = 0;
	long j, seed, *counts = NULL, n = 1, r, repeats = 1, ratio = 0;
	double val = 0.0, weightsum = 0.0, *weights = NULL;
	char *delimstr = "\t";
	const char *errstr;
	struct timeval t1;

	if (pledge("stdio", NULL) == -1)
		err(2, "pledge");

	while ((ch = getopt(argc, argv, "d:nN:r:Rp:s:")) != -1) {
		switch (ch) {
		case 'd':
			delimstr = optarg;
			break;
		case 'n':
			finalnl = 0;
			break;
		case 'N':
			n = strtonum(optarg, 0, LONG_MAX, &errstr);
			if (errstr != NULL)
				errx(1, "bad num value, %s: %s", errstr, optarg);
			break;
		case 'r':
			repeats = strtonum(optarg, 0, LONG_MAX, &errstr);
			if (errstr != NULL)
				errx(1, "bad repeats value, %s: %s", errstr, optarg);
			break;
		case 'R':
			ratio = 1;
			break;
		case 'p':
			prec = strtonum(optarg, 0, INT_MAX, &errstr);
			if (errstr != NULL)
				errx(1, "bad precision value, %s: %s", errstr, optarg);
			break;
		case 's':
			seed = strtonum(optarg, 0, INT_MAX, &errstr);
			if (errstr != NULL)
				errx(1, "bad seed value, %s: %s", errstr, optarg);
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		usage();

	nbins = argc;
	if (!(weights = calloc(nbins, sizeof(double))) ||
	    !(counts = calloc(nbins, sizeof(long))))
		err(1, "calloc");

	for (i = 0; i < nbins; i++) {
		if (!sscanf(argv[i], "%lf", &weights[i]))
			errx(1, "bad weight value: %s", argv[i]);
		if (weights[i] <= 0.0)
			errx(1, "weight %d is not positive (%g)", i, weights[i]);
		if (weights[i] > 1.0)
			errx(1, "weight %d is greater than 1 (%g)", i, weights[i]);
	}

	for (i = 0; i < nbins; i++)
		weightsum += weights[i];
	if (fabs(weightsum - 1.0) > 1e-3)
		errx(1, "weights do not sum to 1 (%g)", weightsum);

	if (s)
#ifdef __OpenBSD__
		srand48_deterministic(seed);
#else
		srand48(seed);
	else {
		gettimeofday(&t1, NULL);
		srand48(t1.tv_sec * t1.tv_usec); /* Caveat: identical seed for each ms */
	}
#endif

	for (r = 0; r < repeats; r++) {
		for (i = 0; i < nbins; i++)
			counts[i] = 0;
		for (j = 0; j < n; j++) {
			val = drand48();
			weightsum = 0.0;
			for (i = 0; i < nbins; i++) {
				weightsum += weights[i];
				if (val <= weightsum) {
					counts[i]++;
					break;
				}
			}
		}
		for (i = 0; i < nbins; i++) {
			if (ratio)
				printf("%.*g", prec, (double)counts[i] / n);
			else
				printf("%ld", counts[i]);
			if (i < nbins - 1)
				fputs(delimstr, stdout);
			else
				if (finalnl)
					putchar('\n');
		}
	}

	free(counts);
	free(weights);

	return 0;
}
